# coding: utf-8
"""Functions and helpers related to guessing configuration attributes."""


from __future__ import unicode_literals

import os
import re
import codecs
from collections import namedtuple
from collections import OrderedDict

from .constants import INPUT


def python_modules():
    """Determine if there are python modules in the cwd.

    Returns:
        list of python modules as strings
    """

    ignored = ["setup.py", "conftest.py"]

    py_modules = []
    for file_ in os.listdir(os.path.abspath(os.curdir)):
        if file_ in ignored or not os.path.isfile(file_):
            continue

        file_name, file_ext = os.path.splitext(file_)
        if file_ext == ".py":
            py_modules.append(file_name)

    return sorted(py_modules)


def potential_data_files(scripts):
    """Determine if there are any potential data files down from cwd.

    Tries to ignore normal things that would pop up you'd like not to include.

    Args:
        scripts: list of known script exectuables

    Returns:
        list of files as relative paths down from cwd
    """

    potential_files = []

    for root, directories, files in os.walk(os.curdir):
        if _ignored(root, is_file=False):
            continue

        for file_ in files:
            if not _ignored(file_):
                relative_path = os.path.join(root, file_)
                while relative_path[0] in (os.curdir, os.path.sep):
                    relative_path = relative_path[1:]
                if relative_path not in scripts:
                    potential_files.append(relative_path)

    return potential_files


def latest_git_tag():
    """Returns the latest tag from git, if .git exists."""

    def git_tag_refs(file_=""):
        return os.path.join(".git", "refs", "tags", file_)

    try:
        git_tags = os.listdir(git_tag_refs())
    except OSError:
        return None

    latest_tag = {0: None}
    for tag in git_tags:
        latest_tag[os.stat(git_tag_refs(tag)).st_ctime] = tag

    return latest_tag[max(latest_tag)]


def find_in_files():
    """Look through most files to try to determine the version and author.

    Returns:
        OrderedDict with the some/all of the following keys if found::

            version, author, author_email, maintainer, maintainer_email
    """

    Guess = namedtuple("Guess", ("source", "weight", "guess"))

    versions = []
    authors = []
    emails = []
    maintainers = []
    maintainer_emails = []

    git_tag = latest_git_tag()
    if git_tag:
        versions.append(Guess("git tag", 99, git_tag))

    to_find = OrderedDict([
        ("version", versions),
        ("author", authors),
        ("email", emails),
        ("author_email", emails),
        ("maintainer", maintainers),
        ("maintainer_email", maintainer_emails),
    ])

    file_weights = {  # additional to the pattern weights
        "__init__.py": 100,
        "__version__.py": 125,
        "version.py": 75,
    }

    for root, directories, files in os.walk(os.curdir):

        if _ignored(root, is_file=False):
            continue

        for file_ in files:
            file_path = os.path.join(root, file_)
            # ignore large files, allows us to be faster with regex's/reading
            if os.stat(file_path).st_size > 102400:
                continue

            with open(file_path, "rb") as openfile:
                try:
                    content = codecs.decode(openfile.read(), "utf-8")
                except UnicodeDecodeError:
                    continue

            for name, guesses in to_find.items():
                for i in range(3):
                    pattern = r'^{u}{name}{u} *= *[\'"]([^\'"]*)[\'"]'.format(
                        u="_" * i,
                        name=name,
                    )

                    re_match = re.search(pattern, content, re.M)
                    if re_match:
                        match_weight = 25 * (i + 1)  # more _ == more important
                        if file_ in file_weights:
                            match_weight += file_weights[file_]

                        guess = re_match.group(1)
                        try:
                            # try to use an ascii string if possible
                            guess = str(codecs.encode(guess, "ascii").decode())
                        except UnicodeEncodeError:
                            pass

                        guesses.append(Guess(
                            "{u_}{name}{u_} from file {fname}".format(
                                u_="_" * i,
                                name=name,
                                fname=file_path,
                            ),
                            match_weight,
                            guess,
                        ))
                        break

    to_find.pop("email")  # duplicated to match either email or author_email
    return OrderedDict((name, max(guesses, key=lambda x: x.weight).guess) for
                       name, guesses in to_find.items() if guesses)


def _ignored(file_or_dir, is_file=True, _recurse=False):
    """Return a boolean of if this file or dir is ignored.

    Args::

        file_or_dir: string file or directory name
        is_file: boolean of if this is a file
    """

    ignored_files = [
        "MANIFEST.in",
        "setup.cfg",
        ".travis.yml",
        ".coverage",
        ".coveragerc",
        ".gitignore",
        "tox.ini",
        "conftest.py",
        "pypackage.meta",
    ]

    ignored_file_patterns = [
        r".*\.pyc$",
        r".*\.py$",  # These will be found as part of a package
    ]

    ignored_dirs = [
        "__pycache__",
        "build",
        "dist",
        ".git",
        ".svn",
        ".eggs",
        "EGG-INFO",
    ]

    ignored_dir_patterns = [
        r"docs?",
        r"examples?",
        r".*\.egg$",
        r".*\.egg-info$",
        r"tests?",
        r".?venv",
    ]

    if not is_file:
        return any([_ignored(part, _recurse=True) for
                    part in file_or_dir.split(os.path.sep)])

    file_or_dir = os.path.basename(file_or_dir)
    ignored = ignored_dirs if _recurse else ignored_files

    if file_or_dir in ignored:
        return True

    patterns = ignored_dir_patterns if _recurse else ignored_file_patterns
    for pattern in patterns:
        if re.match(pattern, file_or_dir, re.I):
            return True

    return False


def _guess_at_things(config):
    """Guesses at attributes of this package's setup configuration.

    Returns:
        OrderedDict of {attribute: guess}
    """

    guesses = OrderedDict(
        name=os.path.basename(os.path.realpath(os.path.curdir))
    )
    guesses.update(find_in_files())

    py_modules = python_modules()
    if py_modules:
        guesses["py_modules"] = py_modules

    scripts = []
    for potential in ("scripts", "bin"):
        root = os.path.join(os.path.abspath(os.curdir), potential)
        if os.path.isdir(root):
            for file_ in os.listdir(root):
                # windows does not have a firm grasp of executable files
                if os.name == "nt" or \
                   os.access(os.path.join(root, file_), os.X_OK):
                    scripts.append(os.path.join(potential, file_))

    if scripts:
        guesses["scripts"] = scripts

    package_files = potential_data_files(scripts)
    if package_files:
        guesses["package_data"] = package_files

    return guesses


def _interactive_confirm(guesses):
    """Prompt the user to confirm or remove guessed attributes."""

    print((
        "pypackage has guessed the following attributes:\n{}\n"
        "would you like to disregard any of the above guesses?"
    ).format("\n".join(['{} `{}` = "{}"'.format(index, attr, guesses[attr])
                       for index, attr in enumerate(guesses, 1)])))

    len_g = len(guesses)
    query = 'use 1-{} or "all" to ignore (enter to accept): '.format(len_g)
    from_user = True

    while from_user and any([guesses[key] for key in guesses]):
        try:
            from_user = INPUT(query)
        except (EOFError, KeyboardInterrupt):
            raise SystemExit("\nInterrupted")
        try:
            from_user = int(from_user)
            assert 1 <= from_user <= len_g  # no negatives, inside len_g
        except:
            if str(from_user).strip().lower() == "all":
                print("ignoring all guesses")
                guesses = {}
            elif from_user:
                print("{} is invalid.".format(from_user))
        else:
            # this is a hack, and the reason guess is an OrderedDict
            from_user = list(guesses.keys())[from_user - 1]
            guesses[from_user] = None
            print("{} will not be guessed".format(from_user))

    return guesses


def perform_guesswork(config, options):
    """Look for missing attributes and take a guess at them.

    Confirm with the user that the guesses are correct after making them if
    they are in interactive mode, otherwise just display them as found.

    Args::

        config: config object to inspect/set attrributes on
        options: options object for cmd line flags
    """

    guesses = _guess_at_things(config)

    if options.interactive:
        guesses = _interactive_confirm(guesses)

    for attr in guesses:
        if (not hasattr(config, attr) or options.re_probe) and guesses[attr]:
            config._metadata_exclusions.append(attr)
            if attr == "package_data":  # set the name as late as possible
                setattr(config, attr, {
                    getattr(config, "name", guesses["name"]): guesses[attr]
                })
            else:
                setattr(config, attr, guesses[attr])

    if getattr(config, "package_data", None) and \
            not getattr(config, "include_package_data", None):
        config.include_package_data = True
        config._metadata_exclusions.append("include_package_data")

    # convert to kwargs here to have config run find_packages
    kwargs = config._as_kwargs

    has_data = getattr(config, "package_data", None)

    if has_data and "packages" not in kwargs and "py_modules" not in guesses:
        # we have no python to distribute, only some binary, move package_data
        config.data_files = [("", list(set(*config.package_data.values())))]
        config._metadata_exclusions.append("data_files")
        delattr(config, "package_data")



"""Tests for the classifier selection process."""


import os
import mock
import pytest
import requests

from pypackage.config import Config

try:
    from pypackage import classified
    HAS_CURSES = True
except ImportError:
    HAS_CURSES = False


# this skips all tests in this module if the client doesn't have curses
pytestmark = pytest.mark.skipif(not HAS_CURSES, reason="no curses support")


def test_choose_classifiers(reset_sys_argv, simple_module):
    """Verify the logic in choosing classifiers."""

    os.environ.pop("ESCDELAY", None)
    conf = Config()
    box = mock.Mock()
    box.pick = mock.Mock(side_effect=iter([0, 4, 4, 4, 0, 9, 4, 3, 0, None]))

    with mock.patch.object(classified, "curses"):
        with mock.patch.object(classified, "BoxSelector", return_value=box):
            choices = classified.choose_classifiers(conf)

    assert choices == ["Development Status :: 4 - Beta"]
    assert "ESCDELAY" not in os.environ


def test_old_escdelay_persists(reset_sys_argv, simple_module):
    """Ensure any previously existing ESCDELAY envvar is reset on exit."""

    os.environ["ESCDELAY"] = "100"
    conf = Config()
    box = mock.Mock()
    box.pick = mock.Mock(return_value=None)

    with mock.patch.object(classified, "curses"):
        with mock.patch.object(classified, "BoxSelector", return_value=box):
            classified.choose_classifiers(conf)

    assert os.environ["ESCDELAY"] == "100"


def test_back_it_up():
    """Ensure we navigate the quasi singly linked list correctly."""

    all_classifiers = classified.read_classifiers()
    parent = all_classifiers.classifiers[8].classifiers[2]
    child = parent.classifiers[1]
    assert classified.back_it_up(child, all_classifiers) == parent


def test_classifiers_current():
    """Ensure the shipped classifiers file is synced with pypi."""

    packaged_trove_file = os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        "pypackage",
        "classifiers"
    )

    with open(packaged_trove_file) as opent:
        packaged = [t.strip() for t in opent.read().splitlines() if t]

    url = "https://pypi.python.org/pypi?:action=list_classifiers"
    trove_req = requests.get(url).text

    current = [t.strip() for t in trove_req.splitlines() if t]

    assert packaged == current, "curl {} > pypackage/classifiers".format(url)


if __name__ == "__main__":
    pytest.main(["-rx", "-v", "--pdb", __file__])



"""Reads static configuration metadata and creates a Python object."""


from __future__ import print_function

import io
import os
import re
import copy
import json
import nose
import yaml
import pytest
import logging
import unittest
from pprint import pformat
from collections import OrderedDict
from setuptools import find_packages

from .runner import TestRunner
from .runner import NOSE_TEMPLATE
from .runner import PYTEST_TEMPLATE
from .runner import UNITTEST_TEMPLATE
from .constants import META_NAME
from .constants import STRING_TYPE
from .constants import VERSION
from .constants import UNICODE


class UNDEF(object):
    """Used to differentiate None/0/False from an option not being defined."""

    pass


class SetOnce(list):
    """Subclass list to have append only append if the value is not in list."""
    def append(self, value):
        if value not in self:
            super(SetOnce, self).append(value)


class Metadata(object):
    """Metadata object class to shim the early stages of PKG-INFO writing.

    Initialized with a Config instance
    """

    def __init__(self, config):
        undef = "UNKNOWN"
        self.metadata_version = "2.0"
        self.generator = "pypackage ({})".format(VERSION)
        self.name = getattr(config, "name", undef)
        self.version = getattr(config, "version", undef)
        self.summary = getattr(config, "description", undef)
        self.description = self.summary
        self.home_page = getattr(config, "url", undef)
        self.url = self.home_page
        self.author = getattr(config, "author", undef)
        self.author_email = getattr(config, "author_email", undef)
        self.license = getattr(config, "license", undef)
        self.platforms = getattr(config, "platforms", undef)
        self.platform = self.platforms  # for PKG-INFO
        self.keywords = getattr(config, "keywords", [])
        self.source_label = getattr(config, "source_label", undef)
        self.source_url = getattr(config, "source_url", undef)

    def write_pkg_info(self, egg_info):
        """Include pypackage custom metadata in PKG-INFO."""

        pkg_info = os.path.join(egg_info, "PKG-INFO")
        info_attrs = ["name", "version", "summary", "home_page",
                      "author", "author_email", "license",
                      "description", "platform", "source_label",
                      "source_url"]
        with io.open(pkg_info, "w", encoding="utf-8") as openinfo:
            openinfo.write(UNICODE("Metadata-Version: 1.0\n"))
            for attr in info_attrs:
                openinfo.write(UNICODE("{}{}: {}\n".format(
                    attr[0].upper(),
                    "".join(attr[1:]).replace("_", "-"),
                    getattr(self, attr),
                )))

    def __str__(self):
        """Self-representation for inclusion in the setup.py."""

        return '''
class Metadata(object):
    """Metadata object class to shim the early stages of PKG-INFO writing."""

    def __init__(self):
        self.metadata_version = {self.metadata_version!r}
        self.generator = {self.generator!r}
        self.name = {self.name!r}
        self.version = {self.version!r}
        self.summary = {self.summary!r}
        self.description = {self.description!r}
        self.home_page = {self.home_page!r}
        self.url = {self.url!r}
        self.author = {self.author!r}
        self.author_email = {self.author_email!r}
        self.license = {self.license!r}
        self.platforms = {self.platforms!r}
        self.platform = {self.platform!r}
        self.keywords = {self.keywords!r}
        self.source_label = {self.source_label!r}
        self.source_url = {self.source_url!r}

    def write_pkg_info(self, egg_info):
        """Include pypackage custom metadata in PKG-INFO."""

        if sys.version_info > (3,):
            UNICODE = str
        else:
            UNICODE = unicode

        pkg_info = os.path.join(egg_info, "PKG-INFO")
        info_attrs = ["name", "version", "summary", "home_page",
                      "author", "author_email", "license",
                      "description", "platform", "source_label",
                      "source_url"]
        with io.open(pkg_info, "w", encoding="utf-8") as openinfo:
            openinfo.write(UNICODE("Metadata-Version: 1.0\\n"))
            for attr in info_attrs:
                openinfo.write(UNICODE("{{}}{{}}: {{}}\\n".format(
                    attr[0].upper(),
                    "".join(attr[1:]).replace("_", "-"),
                    getattr(self, attr),
                )))
'''.format(self=self)


class Config(object):
    """Config object. Attributes are passed as kwargs."""

    _KEYS = OrderedDict([  # all available configuration keys: value type
        ("name", str),
        ("version", str),
        ("description", str),
        ("long_description", str),
        ("author", str),
        ("author_email", str),
        ("maintainer", str),
        ("maintainer_email", str),
        ("url", str),
        ("download_url", str),
        ("packages", list),
        ("py_modules", list),
        ("scripts", list),
        ("entry_points", {str: list}),
        ("install_requires", (str, list)),
        ("tests_require", (str, list)),
        ("classifiers", list),
        # ~~ separation of standard/extended here ~~
        ("ext_modules", list),
        ("keywords", list),
        ("platforms", list),
        ("cmdclass", {str: object}),
        ("data_files", (list,)),
        ("package_dir", {str: str}),
        ("build_requires", (str, list)),
        ("setup_requires", (str, list)),
        ("extras_require", {str: list}),
        ("include_package_data", bool),
        ("exclude_package_data", {str: list}),
        ("package_data", {str: list}),
        ("zip_safe", bool),
        ("dependency_links", list),
        ("namespace_packages", list),
        ("test_suite", str),
        ("test_loader", str),
        ("eager_resources", list),
        ("use_2to3", bool),
        ("convert_2to3_doctests", list),
        ("use_2to3_fixtures", list),
        ("extensions", {str: list}),
    ])
    # our distinct keys, these do not get passed to setuptools/distutils
    _PYPACKAGE_KEYS = OrderedDict([
        ("test_runner", str),
        ("tests_dir", str),     # no function required for these two, they're
        ("runner_args", list),  # only used in the enable_test_runner function
        ("source_label", str),  # these two are in a draft pep 426
        ("source_url", str),    # early implementation here
    ])

    def __init__(self, **kwargs):
        """Builds a Config object, fills in options passed as kwargs."""

        config_as_dict = {key: UNDEF() for key in Config._KEYS}
        config_as_dict.update({key: UNDEF() for key in Config._PYPACKAGE_KEYS})
        self._defaults = site_defaults()
        config_as_dict.update(self._defaults)
        config_as_dict.update(kwargs)

        extras = {}
        for key, value in config_as_dict.items():
            if not isinstance(value, UNDEF):
                if re.match(".*_requires?$", key) and key not in Config._KEYS:
                    extras[key.rpartition("_")[0]] = value
                else:
                    setattr(self, key, value)

        if extras:
            if hasattr(self, "extras_require"):
                self.extras_require.update(extras)
            else:
                self.extras_require = extras

        # duplicate in some values
        duplicate_values = [
            ("description", "long_description"),
            ("author", "maintainer"),
            ("author_email", "maintainer_email"),
        ]
        for orig, alias in duplicate_values:
            if hasattr(self, orig):
                setattr(self, alias, getattr(self, alias, getattr(self, orig)))
                if getattr(self, orig) == getattr(self, alias):
                    self._defaults[alias] = getattr(self, alias)

        # _verify will toggle these if set
        self._configured_runner_args = False
        self._configured_tests_require = False
        self._long_read = ""  # used in direct setuptools interactions
        self._long_read_in_setup = ""  # used in writing the setup.py

        # filled in during guessing phase
        self._metadata_exclusions = SetOnce()

        # perform init-time type validation and runs feature functions
        self._verify()

    @property
    def _as_kwargs(self):
        """Builds a dict suitable for use with setuptools.setup directly."""

        kwargs = OrderedDict()

        for key in Config._KEYS:
            if hasattr(self, key):
                if key == "packages":
                    kwargs[key] = self._packages_actual()
                elif key == "long_description":
                    kwargs[key] = self._long_read or self.long_description
                else:
                    kwargs[key] = getattr(self, key)

        if "packages" not in kwargs:
            packages = find_packages(exclude=["test", "tests"])
            if packages:
                kwargs["packages"] = packages

        if hasattr(self, "metadata"):
            kwargs["metadata"] = self.metadata

        return kwargs

    @property
    def _metadata(self):
        """Metadata unique to this config."""

        metadata = OrderedDict([(k, v) for k, v in self._as_kwargs.items()
                                if k not in self._defaults])
        # remove the cmdclass key and tests_require if not set by the user
        metadata.pop("cmdclass", None)
        if not self._configured_tests_require:
            metadata.pop("tests_require", None)

        # remove anything that's been guessed
        for key in self._metadata_exclusions:
            metadata.pop(key, None)

        # reset long_description back to the filename
        if self._long_read:
            metadata["long_description"] = self.long_description

        # add in feature keys that have been set
        for attr in Config._PYPACKAGE_KEYS:
            if hasattr(self, attr):
                if attr != "runner_args" or self._configured_runner_args:
                    metadata[attr] = getattr(self, attr)

        return metadata

    def __str__(self):
        """Creates a string representation of self, resulting in a setup.py."""

        packages_str, find_needed = self._packages_string()
        cmdclass = self._cmdclass_string()
        long_descr_str = self._long_description_string()

        ignored = ["metadata"]
        altered_keys = ["packages", "long_description", "cmdclass"]
        altered_strs = [packages_str, long_descr_str, cmdclass]

        imports = ["from setuptools import setup"]
        if find_needed:
            imports.append("from setuptools import find_packages")
        if self._long_read_in_setup:
            imports.insert(0, "import io")
        if hasattr(self, "metadata"):
            if "import io" not in imports:
                imports.insert(0, "import io")
            imports.insert(1, "import sys")
            imports.insert(1, "import os")
            altered_keys.append("metadata")
            altered_strs.append("metadata=Metadata()")

        return "\n".join([
            '"""{}\'s setup.py.\n'.format(
                getattr(self, "name", "None").title()
            ),
            "This file was autogenerated by pypackage. To make changes, you",
            "should edit the {} rather than this setup.py.".format(META_NAME),
            '"""\n\n',
            "\n".join(imports),
            self._test_runner_string() or "",
            str(self.metadata) if hasattr(self, "metadata") else "",
            "{}setup(".format(self._long_read_in_setup),
            "\n".join([
                "    {}={},".format(key, _multiline(val)) for key, val in
                self._as_kwargs.items() if key not in altered_keys + ignored
            ]),
            "\n".join(
                ["    {},".format(altered_str) for altered_str in altered_strs
                    if altered_str]
            ) or "",
            ")",
        ])

    def _long_description_string(self):
        """Builds a string for long_description (+file read) in setup.py."""

        if not hasattr(self, "long_description"):
            return

        if self._long_read_in_setup:
            return "long_description=long_description"
        else:
            return "long_description={!r:}".format(self.long_description)

    def _test_runner_string(self):
        """Builds a string from the test_runner template, if in use."""

        if not hasattr(self, "test_runner"):
            return

        if self.test_runner == "unittest":  # extra formatting
            return UNITTEST_TEMPLATE.format(
                self=self,
                tests_dir=getattr(self, "tests_dir", "."),
            )
        elif self.test_runner == "nose":
            return NOSE_TEMPLATE.format(self=self)
        else:
            return PYTEST_TEMPLATE.format(self=self)

    def _packages_actual(self):
        """Returns the list of packages for direct use with setuptools."""

        for package in self.packages:
            if package.startswith("find_packages(") and \
                    ";" not in package and "." not in package:
                # no ;'s or .'s here to reduce attack surface
                try:
                    found_packages = eval(package)
                except:
                    continue
                else:
                    return found_packages
        else:
            return self.packages

    def _packages_string(self):
        """Builds a string for `packages=` in the setup.py.

        Returns:
            tuple of a packages string and boolean of requiring find_packages
        """

        # defaults
        excludes = ["test", "tests"]
        package_str = "find_packages(exclude={!r:})".format(excludes)
        find_required = True

        if hasattr(self, "packages"):
            for package in self.packages:
                if "find_packages(" in package:
                    # user supplied a find_packages() string
                    package_str = package
                    break
            else:
                # user supplied packages list
                package_str = repr(self.packages)
                find_required = False

        return "packages={}".format(package_str), find_required

    def _cmdclass_string(self):
        """Builds a string for `cmdclass=` in the setup.py."""

        cmdclass = getattr(self, "cmdclass", {})
        as_string = copy.deepcopy(cmdclass)
        if "test" in as_string:
            as_string["test"] = "PyPackageTest"  # name in template

        if as_string:
            return "cmdclass={{{}}}".format(", ".join(  # repr the keys only
                "{!r:}: {}".format(k, v) for k, v in as_string.items()
            ))

    def __repr__(self):
        return "<{name}.{cls} object at {id}, {attributes}>".format(
            name=__name__,
            cls=self.__class__.__name__,
            id=hex(id(self)),
            attributes=", ".join(
                ["{attr}: '{{self.{attr}}}'".format(attr=attr)
                 for attr in dir(self) if not attr.startswith("_")]
            ).format(self=self),
        )

    def _enable_long_description_read(self):
        """Sets up the automatic read feature of long_description.

        Returns:
            None, modifies attributes of self
        """

        if self._long_read or not hasattr(self, "long_description"):
            return

        try:
            if os.path.isfile(self.long_description):
                with io.open(self.long_description, encoding="utf-8") as descr:
                    self._long_read = descr.read()
                self._long_read_in_setup = (
                    'with io.open("{}", encoding="utf-8") as opendescr:\n'
                    "    long_description=opendescr.read()\n\n\n"
                ).format(
                    self.long_description
                )
            else:
                return
        except:
            return

    def _enable_test_runner(self):
        """Sets up the preferred testing suite.

        Returns:
            None, modifies attributes of self
        """

        if not hasattr(self, "test_runner"):
            return

        test_runner = self.test_runner.lower()
        if test_runner == "pytest":
            self._enable_pytest()
        elif test_runner.startswith("nose"):  # allow nosetests... etc
            self.test_runner = "nose"  # exact name for importing though
            self._enable_nosetest()
        else:
            self.test_runner = "unittest"
            self._enable_unittest()

        TestRunner.pypackage(self)  # XXX after runner_args are set
        self.cmdclass = {"test": TestRunner}

    def _enable_nosetest(self):
        """Do nosetest specific logic to enable it as a test runner."""

        default_args = ["-v", "-d", "--with-coverage", "--cov-report",
                        "term-missing", "--cov"]

        self._runner = nose

        # grab the user's tests_require, make sure nose is in there
        self.tests_require = getattr(self, "tests_require", None)
        if self.tests_require is None:
            self.tests_require = ["nose"]
        else:
            self._configured_tests_require = self.tests_require not in (
                ["nose"], ["nose", "nose-cov"])
            if "nose" not in self.tests_require:
                self.tests_require.append("nose")

        # configure the default or user supplied runner arguments
        arg_len = 1 + (int(hasattr(self, "tests_dir")) * 2)  # *2 b/c -w flag
        self.runner_args = getattr(self, "runner_args", None)
        if self.runner_args is None:
            self.runner_args = default_args[:2]
            if hasattr(self, "name"):
                self.runner_args.extend(default_args[2:] + [self.name])
                if "nose-cov" not in self.tests_require:
                    self.tests_require.append("nose-cov")
        elif len(self.runner_args) == len(default_args) + arg_len and \
                self.runner_args[:-arg_len] == default_args:
            # refresh runner_args in case our name has changed for coverage
            self.runner_args = default_args + [getattr(self, "name", "")]
            if "nose-cov" not in self.tests_require:
                self.tests_require.append("nose-cov")
        else:
            self._configured_runner_args = True  # include them in metadata

        # use -w to specify NOSEWHERE, or let nose find the tests itself
        if hasattr(self, "tests_dir"):
            self.runner_args.extend(["-w", self.tests_dir])

    def _enable_unittest(self):
        """Do unittest specific logic to enable it as a test runner."""

        self._runner = unittest
        self.runner_args = getattr(self, "runner_args", None)
        if self.runner_args is None:
            self.runner_args = []
        else:
            self._configured_runner_args = True

    def _enable_pytest(self):
        """Do pytest specific logic to enable it as a test runner."""

        default_args = ["-v", "-rx", "--cov-report", "term-missing", "--cov"]

        self._runner = pytest

        # grab the user's tests_require, make sure pytest is in there
        self.tests_require = getattr(self, "tests_require", None)
        if self.tests_require is None:
            self.tests_require = ["pytest"]
        else:
            self._configured_tests_require = self.tests_require not in (
                ["pytest"], ["pytest", "pytest-cov"])
            if "pytest" not in self.tests_require:
                self.tests_require.append("pytest")

        # configure the default or user supplied runner arguments
        arg_len = 1 + int(hasattr(self, "tests_dir"))  # name + dir if supplied
        self.runner_args = getattr(self, "runner_args", None)
        if self.runner_args is None:
            self.runner_args = default_args[:2]
            if hasattr(self, "name"):
                self.runner_args.extend(default_args[2:] + [self.name])
                if "pytest-cov" not in self.tests_require:
                    self.tests_require.append("pytest-cov")
        elif len(self.runner_args) == len(default_args) + arg_len and \
                self.runner_args[:-arg_len] == default_args:
            # refresh runner_args in case our name has changed for coverage
            self.runner_args = default_args + [getattr(self, "name", "")]
            if "pytest-cov" not in self.tests_require:
                self.tests_require.append("pytest-cov")
        else:
            self._configured_runner_args = True  # include them in metadata

        # tack the tests dir on the end, or let pytest find them
        if hasattr(self, "tests_dir"):
            self.runner_args.append(self.tests_dir)

    def _enable_metadata_obj(self):
        """If source_label or source_url is used, build the metadata object."""

        if any(getattr(self, k, None) for k in ("source_url", "source_label")):
            self.metadata = Metadata(self)

    def _verify(self):
        """Ensures self attributes conform to their type declarations."""

        self._enable_test_runner()
        self._enable_long_description_read()
        self._enable_metadata_obj()

        for key, type_ in list(Config._KEYS.items()) + list(
                Config._PYPACKAGE_KEYS.items()):
            if hasattr(self, key):
                self._verify_key(key, type_)

    def _verify_key(self, key, type_):
        """Verify that key is of type type_.

        Raises:
            TypeError if they key cannot be coerced into type_
        """

        if isinstance(type_, dict) and isinstance(getattr(self, key), dict):
            setattr(self, key, ensure_dict(getattr(self, key), type_))
        elif isinstance(type_, dict):
            raise TypeError("{} should be a dict, not {}!".format(
                key,
                type(getattr(self, key)).__name__,
            ))
        elif type_ is list and isinstance(getattr(self, key), list):
            setattr(self, key, ensure_list(getattr(self, key)))
        elif type_ is list:
            setattr(self, key, [getattr(self, key)])
        elif not isinstance(getattr(self, key), type_):
            if isinstance(type_, tuple):  # multiple acceptable values
                for type__ in type_:
                    if type__ is list:
                        setattr(self, key, [getattr(self, key)])
                        break
                    else:
                        try:
                            setattr(self, key, type__(getattr(self, key)))
                            break
                        except:
                            pass
                else:
                    raise TypeError("{} should be a {} or {}, not {}!".format(
                        key,
                        ", ".join([t.__name__ for t in type_[:-1]]),
                        type_[-1].__name__,
                        type(getattr(self, key)).__name__,
                    ))
            else:
                try:
                    setattr(self, key, type_(getattr(self, key)))
                except:
                    raise TypeError("{} should be a {}, not {}!".format(
                        key, type_.__name__, type(getattr(self, key)).__name__,
                    ))


def _multiline(value, indent=4):
    """Return value as a multiline (pretty) string, with indent."""

    val = pformat(value)
    if "\n" in val:
        value_as_lines = val.splitlines()

        indented = [
            "{}".format(value_as_lines[0][0]),  # rip the leading bracket
            "{}{}".format(" " * (indent * 2), value_as_lines[0][1:].strip()),
        ]
        for line in value_as_lines[1:-1]:
            indented.append("{}{}".format(" " * (indent * 2), line.strip()))
        indented.extend([
            "{}{}".format(" " * (indent * 2), value_as_lines[-1][:-1].strip()),
            "{}{}".format(" " * indent, value_as_lines[-1][-1]),
        ])
        val = "\n".join(indented)
    return val


def ensure_list(list_to_verify):
    """Ensures that all items in the list are of string type.

    Also guarentees that all items in the list are unique. Maintains order.
    """

    string_list = []

    def string_list_append(item):
        if item not in string_list:
            string_list.append(item)

    for item in list_to_verify:
        if isinstance(item, STRING_TYPE):
            string_list_append(item)
        else:
            string_list_append(str(item))

    return string_list


def ensure_dict(dict_to_verify, dict_types):
    """Goes through dict_to_verify and tries to coerce into dict_types.

    Args::

        dict_to_verify: a dict with keys and values to verify
        dict_types: a dict with a single key/value pair of the expected types

    Returns:
        a new dictionary, coerced into the correct types
    """

    key_type = list(dict_types.keys())[0]
    value_type = list(dict_types.values())[0]
    return_dict = {}
    for key, value in dict_to_verify.items():
        if not isinstance(key, key_type):
            key = key_type(key)  # this will raise on error...
        if not isinstance(value, value_type):
            if value_type is list and isinstance(key, STRING_TYPE):
                value = [value]  # auto list single string values
            else:
                value = value_type(value)  # this will also raise on error...
        return_dict[key] = value
    return return_dict


def site_defaults():
    """Look for site default metadata to mixin onto any missing values."""

    filename = os.path.join(os.path.expanduser("~"), ".pypackage")
    if os.path.isfile(filename):
        return json_maybe_commented(filename) or {}
    else:
        logging.debug("Site defaults requested but not found at %s", filename)
        return {}


def get_config(path=None):
    """Looks in path or os.cwd for a static metadata file and builds a config.

    Args:
        path: optional string path to search inside, or None for cwd

    Returns:
        Config object to build/install with or None on error
    """

    if path is None:
        path = os.path.abspath(os.path.curdir)

    pyjson = os.path.join(path, META_NAME)
    if os.path.isfile(pyjson):
        return Config(**json_maybe_commented(pyjson))
    else:
        logging.info("Using site defaults, no %s found in %s", META_NAME, path)
        return Config()


def reduce_json_unicode(json_obj):
    """Stupidly dumps JSON into YAML to remove what unicode it can."""

    return yaml.safe_load(json.dumps(json_obj))


def json_maybe_commented(filename, remove_comments=False):
    """Tries twice to load the filename as JSON.

    The second time, if needed, it will remove lines starting with `#`.
    """

    if remove_comments:
        cleaned_lines = []
        offset = 0
        with open(filename, "r") as openfile:
            for line in openfile:
                if re.match("\s*#", line):  # leading whitespace then #
                    offset += 1
                else:
                    cleaned_lines.append(line)
        try:
            return reduce_json_unicode(json.loads("".join(cleaned_lines)))
        except Exception as error:
            meta_line = ""
            if offset:
                if hasattr(error, "lineno"):  # new python
                    meta_line = " (originally line {} in .meta)".format(
                        offset + error.lineno
                    )
                else:  # old python
                    try:
                        meta_line = " (originally line {} in .meta)".format(
                            offset +
                            int(error.args[0].split("line ")[1].split()[0])
                        )
                    except:
                        pass

            raise SystemExit("Error reading json from {}: {!r}{}".format(
                filename,
                error,
                meta_line,
            ))
    else:
        try:
            with open(filename, "r") as openfile:
                return reduce_json_unicode(json.loads(openfile.read()))
        except Exception as error:
            logging.debug("Commented JSON? Recursing because: %r", error)
            return json_maybe_commented(filename, remove_comments=True)



"""Entry points for running setup.py commands.

Each command's docstring is used as additional help if the -h flag is used.

Setup is a bit of a snowflake in that it doesn't build anything, takes no args.

Info does not take any arguments, only looks up packages by name.
"""


from __future__ import print_function

import sys
import pkg_resources

from . import pypackage_setup
from .config import get_config
from .cmdline import get_options
from .cmdline import help_and_version


def install():
    """py-install will build a setup.py and use it to install locally."""

    pypackage_setup(["install"], additional=install.__doc__)


def run_tests():
    """py-test will build a setup.py and run the test command with it.

    It is recommended to supply a 'test_runner' argument in your metadata to
    use this command. Setting it to one of "pytest", "nose", or "unittest" will
    cause this command to execute a test discovery and execution. You can also
    specify the "tests_dir" if you find the test runner is unable to find your
    tests. The "runner_args" key is set per runner to values that include some
    level of verbosity and coverage statistics, but feel free to override that
    value in your packages' metadata file to something more your liking.

    Note that pytest and nosetest do not expect __init__.py files in your test
    directories, whereas unittest requires them to find the tests.\
    """

    pypackage_setup(["test"], additional=run_tests.__doc__)


def develop():
    """py-develop will build a setup.py and run the develop command with it."""

    pypackage_setup(["develop"], additional=develop.__doc__)


def build():
    """py-build will build and run a setup.py from metadata and/or inspection.

    After running py-build you should have a dist directory containing a new
    source release and wheel release which you could then upload to pypi and/or
    install with pip.

    If you only want to build the metadata/setup.py and not the package itself,
    which can be handy during development, use one of the -s or -m flags.\
    """

    options = get_options()
    build_commands = ["build", "sdist", "bdist_wheel"]
    if options.setup or options.metadata:
        # don't run the setup.py, just build it and/or remake the metadata
        build_commands = None
    pypackage_setup(build_commands, options=options, additional=build.__doc__)


@help_and_version
def setup():
    """py-setup will create a setup.py from metadata files and/or inspection.

    It only prints this configuration to stdout, it does not run the setup.py,
    do any guesswork/inspection, or even save it to disk (py-build does that).

    You can use this command to test how changes in your metadata translate to
    changes in the setup.py.

    Usage:
        py-setup [FILEPATH]

    positional arguments:
        FILEPATH    Metadata directory location (default: cwd)\
    """

    if len(sys.argv) >= 2:
        for arg in sys.argv[1:]:
            print(get_config(arg))
    else:
        print(get_config())


@help_and_version
def info():
    """py-info will print the most recent version of a package's metadata.

    Usage:
        py-info <package> [package] ...\
    """

    env = pkg_resources.Environment()
    separator = False
    for arg in sys.argv[1:]:
        try:
            pkg = env[pkg_resources.safe_name(arg)][0]
        except IndexError:
            print(
                "The package {} was not found.".format(arg),
                file=sys.stderr,
            )
            continue

        # this is what a decade of legacy code looks like
        if pkg.PKG_INFO == "METADATA":
            # don't look too close, this is an email message object...
            meta = pkg._parsed_pkg_info.items()
            print("{}{}".format(
                "{}\n".format("-" * 40) if separator else "",
                "\n".join([
                    "{}: {}".format(k, v) for k, v in meta if v != "UNKNOWN"
                ]),
            ))
        else:
            # this one's a string generator
            meta = pkg._get_metadata("PKG-INFO")
            print("{}{}".format(
                "{}\n".format("-" * 40) if separator else "",
                "\n".join([
                    line for line in meta if "UNKNOWN" not in line
                ])
            ))

        separator = True



"""Tests for the install command.

Unfortunately, to make this work with setuptools and pytest and importlib,
we cannot actually install here and import to check for success.

Mock is used to catch and inspect the call to setuptools instead.
"""


import os
import sys
import mock
import pytest
import setuptools

from pypackage.commands import install


def test_simple_module(simple_module):
    """Tests installing a simple python module."""

    with mock.patch.object(setuptools, "setup") as mocked_setup:
        install()

    assert sys.argv == ["setup.py", "install"]
    mod_dir, mod_name = simple_module
    mocked_setup.assert_called_once_with(**{
        "name": os.path.basename(mod_dir),
        "py_modules": [mod_name],
    })


def test_simple_package(simple_package):
    """Tests installing a simple python package."""

    with mock.patch.object(setuptools, "setup") as mocked_setup:
        install()
    assert sys.argv == ["setup.py", "install"]
    mocked_setup.assert_called_once_with(**{
        "name": os.path.basename(simple_package),
        "packages": [os.path.basename(simple_package)],
    })


def test_with_data(with_data):
    """Tests installing a python package with data."""

    with mock.patch.object(setuptools, "setup") as mocked_setup:
        install()
    assert sys.argv == ["setup.py", "install"]
    root, pkg_root = with_data
    package = os.path.basename(pkg_root)
    mocked_setup.assert_called_once_with(**{
        "name": package,
        "packages": [package],
        "package_data": {package: ["{}/data/data_1".format(package)]},
        "include_package_data": True,
        "long_description": "this package has data!",
    })


def test_with_scripts(with_scripts):
    """Tests installing with scripts."""

    with mock.patch.object(setuptools, "setup") as mocked_setup:
        install()
    assert sys.argv == ["setup.py", "install"]
    package_dir, scripts_dir = with_scripts
    package = os.path.basename(package_dir)
    mocked_setup.assert_called_once_with(**{
        "name": package,
        "packages": [package],
        "scripts": ["{}/my_script".format(scripts_dir)],
    })


def test_long_description_as_file(with_readme):
    """Test installing with a filepath as a long_description."""

    with mock.patch.object(setuptools, "setup") as mocked_setup:
        install()
    assert sys.argv == ["setup.py", "install"]
    package_dir, mod_root = with_readme
    package = os.path.basename(mod_root)
    mocked_setup.assert_called_once_with(**{
        "name": package,
        "packages": [package],
        "long_description": "{n}\n{d}\n\n{n}'s readme... with content!".format(
            n=package,
            d="=" * len(package),
        ),
        "package_data": {package: ['README']},
        "include_package_data": True,
    })


def test_source_tags(source_release):
    """Ensure the source url and label are in the package metadata."""

    with mock.patch.object(setuptools, "setup") as mocked_setup:
        install()

    assert sys.argv == ["setup.py", "install"]
    package_dir, source_label, source_url = source_release
    package = os.path.basename(package_dir)
    call_args = mocked_setup.call_args[1]
    assert call_args["name"] == package
    assert call_args["packages"] == [package]
    assert "<pypackage.config" in repr(call_args["metadata"])
    assert ".Metadata object" in repr(call_args["metadata"])
    assert callable(call_args["metadata"].write_pkg_info)


if __name__ == "__main__":
    pytest.main(["-rx", "-v", "--pdb", __file__])



"""Ensure we are parsing arguments corretly."""


import sys
import pytest

from pypackage.cmdline import get_options


def test_get_options(reset_sys_argv):
    sys.argv = ["py-build", "-isR", "--extended"]
    options = get_options()
    for opt in ("interactive", "setup", "re_classify", "extended"):
        assert getattr(options, opt)


if __name__ == "__main__":
    pytest.main(["-rx", "-v", "--pdb", __file__])
